#include "../../include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.h>

#include <stdio.h>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            __declare_io_type(Mode, uint8_t, 
                static constexpr auto DoubleSpeed    = Bit(7);
                static constexpr auto SingleSpeed    = !DoubleSpeed;
                static constexpr auto XADPCM         = Bit(6);
                static constexpr auto WholeSector    = Bit(5);
                static constexpr auto DataSector     = !WholeSector;
                static constexpr auto UseXAFilter    = Bit(3);
                static constexpr auto AudioPlayIRQ   = Bit(2);
                static constexpr auto AutoPauseTrack = Bit(1);
                static constexpr auto CDDA           = Bit(0);
            );

            static constexpr auto DataSectorMode = Mode_t::from(Mode_t::DoubleSpeed, Mode_t::DataSector);
            
            static SectorBufferAllocator sector_allocator;
            static uint16_t              sectors_left;

            static InterruptVerifierResult interrupt_verifier() {
                static const auto pause = []() {
                    CD_IO::PortIndex0::change_to();
                    Command::send<CD_IO::PortIndex0>(CD_IO::Command::Pause);
                };

                if(Interrupt::is_irq(Interrupt::CDROM)) {
                    const uint8_t old_idx = (CD_IO::IndexStatus & 0x3);

                    CD_IO::PortIndex1::change_to();
                    const auto cur_irq = CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag);
                    last_interrupt = cur_irq;
                    CD_IO::Interrupt::ack(CD_IO::PortIndex1::InterruptFlag);

                    if(cur_irq == CD_IO::Interrupt::DataReady) {
                        //Obtain sector content here
                        auto* sector = sector_allocator.allocate_sector();
                        if(sector) {
                            //Now obtain sector
                            static const char SimpleStr[] = "CD-Drive: This is a test string";

                            char* test_str = reinterpret_cast<char*>(sector->data);
                            for(size_t n = 0; n < sizeof(SimpleStr); n++) {
                                test_str[n] = SimpleStr[n];
                            }

                            sectors_left--;
                            if(sectors_left == 0) {
                                current_state = State::Done;
                                pause();
                            }                           
                        }

                        else {
                            current_state = State::BufferFull;
                            pause();
                        }
                    }

                    else if(cur_irq == CD_IO::Interrupt::DiskError) {
                        current_state = State::Error;
                    }

                    CD_IO::IndexStatus = old_idx;
                    return InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return InterruptVerifierResult::SkipHandler;
                }
            }

            static void interrupt_handler(uint32_t) {
                Interrupt::ack_irq(Interrupt::CDROM);
                __syscall_ReturnFromException();
            }

            CD_IO::Interrupt::Type last_interrupt = CD_IO::Interrupt::Type::None;
            State                  current_state  = State::Free;
            InterrupCallback callback = {
                .next              = nullptr,
                .handler_function  = reinterpret_cast<InterruptHandler>(interrupt_handler),
                .verifier_function = interrupt_verifier
            };

            void read_file(FileInfo file_info, const SectorBufferAllocator& buffer_allocator) {
                sector_allocator = buffer_allocator;
                sectors_left     = file_info.sectors;

                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, DataSectorMode);

                const auto loc = CDTimeStamp::from(file_info);
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, loc.get_min_cd(), loc.get_sec_cd(), loc.get_sector_cd());
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::ReadN);

                current_state = State::Reading;

                printf("Now reading: %i\n", file_info.lba);
            }
        }
    }
}